------------------------------------------------------------
--                                                        --
-- FSAL (Full Screen Assistant for LÖVE)                  --
--                                           Version 0.09 --
--                                                        --
-- Copyright (C) 2020-2024 Michiyo Tagami                 --
--                                                        --
-- Released under the MOUSEY license                      --
--                                                        --
------------------------------------------------------------

local fsal_width   -- Horizontal size of screen (px)
local fsal_height  -- Vertical size of screen (px)
local fsal_scalex  -- Horizontal scale
local fsal_scaley  -- Vertical scale
local fsal_scale   -- Fixed scale
local fsal_offsetx -- Horizontal Display offset
local fsal_offsety -- Vertical Display offset
local fsal_cwidth  -- Horizontal size of canvas (px)
local fsal_cheight -- Vertical size of canvas (px)
local fsal_scmode  -- Full screen flag
local fsal_cvone   -- Canvas
local fsal_hvmode  -- Horizontal or Vertical
local fsal_defaultw -- Default width of window
local fsal_defaulth -- Default height of window
local fsal_deskwidth -- Desktop width
local fsal_deskheight -- Desktop height
local fsal_mcmode -- Mouse clipping flag

-- fsal_initialize(width, height, onoff, hidemc, grab)
--   Initialize FSAL
--     width : Horizontal size of canvas (px)
--     height: Vertical size of canvas (px)
--     onoff : Full screen flag
--       true: fullscreen / false: windowed
--     hidemc: Whether or not the mouse cursor is displayed in full screen.
--       true: show mouse cursor / false: hide mouse cursor
--     grab  : Mouse grab flag
--       true: grab mouse cursor / false: free mouse cursor
function fsal_initialize(width, height, onoff, hidemc, grab)
  love.window.setFullscreen(onoff, "desktop")
  fsal_defaultw, fsal_defaulth = width, height
  fsal_width, fsal_height = love.graphics.getDimensions()
  fsal_deskwidth, fsal_deskheight = love.window.getDesktopDimensions()
  --print(fsal_deskwidth, fsal_deskheight)
  --print(fsal_width.." ".. fsal_height)
  fsal_scalex = fsal_width / width
  fsal_scaley = fsal_height / height
  if (fsal_scalex<fsal_scaley) then
    fsal_hvmode = false
    fsal_offsetx = 0
    fsal_offsety = ((fsal_height / fsal_scalex) - height) / 2
    fsal_scale = fsal_scalex
  else
    fsal_hvmode = true
    fsal_offsetx = ((fsal_width / fsal_scaley) - width) / 2
    fsal_offsety = 0
    fsal_scale = fsal_scaley
  end
  fsal_cwidth = width
  fsal_cheight = height
  fsal_scmode = onoff
  fsal_cvone = love.graphics.newCanvas(fsal_cwidth, fsal_cheight)
  --love.window.setFullscreen(onoff, "desktop")
  --if (hidemc and fsal_scmode) then
  if hidemc then
    love.mouse.setVisible(false)
  else
    love.mouse.setVisible(true)
  end
  if (grab~=nil) then
    love.mouse.setGrabbed(grab)
  end
end

-- fsal_modechange(hidemc, grab)
--   Switch between full screen mode and window mode
--     hidemc: Whether or not the mouse cursor is displayed in full screen.
--       true: show mouse cursor / false: hide mouse cursor
--     grab  : Mouse grab flag
--       true: grab mouse cursor / false: free mouse cursor
function fsal_modechange(hidemc, grab)
  fsal_scmode = not fsal_scmode
  love.window.setFullscreen(fsal_scmode, "desktop")
  fsal_width, fsal_height = love.graphics.getDimensions()
  fsal_scalex = fsal_width / fsal_cwidth
  fsal_scaley = fsal_height / fsal_cheight
  if (fsal_scalex<fsal_scaley) then
    fsal_hvmode = false
    fsal_offsetx = 0
    fsal_offsety = ((fsal_height / fsal_scalex) - fsal_cheight) / 2
    fsal_scale = fsal_scalex
  else
    fsal_hvmode = true
    fsal_offsetx = ((fsal_width / fsal_scaley) - fsal_cwidth) / 2
    fsal_offsety = 0
    fsal_scale = fsal_scaley
  end
  if (hidemc) then
    love.mouse.setVisible(false)
  else
    love.mouse.setVisible(true)
  end
  if (grab~=nil) then
    love.mouse.setGrabbed(grab)
  end
end

-- fsal_canvas()
--   Prepare the drawing canvas
function fsal_canvas()
  return fsal_cvone
end

-- fsal_drawstart(canvas)
--   Start drawing
--     canvas: Target canvas (usually prepared with fsal_canvas)
function fsal_drawstart(canvas)
  --love.graphics.scale(1, 1)
  love.graphics.setCanvas(canvas)
  love.graphics.clear()
end

-- fsal_drawend()
--   Ends drawing
function fsal_drawend()
  love.graphics.setCanvas()
end

-- fsal_show(canvas)
--   Display the drawn contents on the screen
--     canvas: Target canvas (usually prepared with fsal_canvas)
function fsal_show(canvas)
  love.graphics.push("all")
  love.graphics.origin()
  --love.graphics.scale(fsal_scale, fsal_scale)
  --love.graphics.setColor(1,1,1,1)
  love.graphics.setColor(love.graphics.getBackgroundColor())
  --love.graphics.rectangle("fill", fsal_offsetx, fsal_offsety, fsal_cwidth, fsal_cheight)
  love.graphics.rectangle("fill", fsal_offsetx*fsal_scale, fsal_offsety*fsal_scale,
    fsal_cwidth*fsal_scale, fsal_cheight*fsal_scale)
  love.graphics.setColor(1,1,1,1)
  love.graphics.draw(canvas, fsal_offsetx*fsal_scale, fsal_offsety*fsal_scale,0,fsal_scale, fsal_scale)
  --print(fsal_scale)
  --love.graphics.scale(1, 1)
  love.graphics.pop()
end

-- fsal_drawbgimage(image, quad, iw, ih)
--   Draw an image on the background visible outside the main screen.
--   Use before fsal_drawstart(canvas).
--     image: Image object to draw in the background
--     quad : Giving a Quad here allows for background animation, etc.
--     iw   : When quad is given, specify the width of one frame.
--     ih   : When quad is given, specify the height of one frame.
function fsal_drawbgimage(image, quad, iw, ih)
  love.graphics.push("all")
  love.graphics.origin()
  love.graphics.setColor(1, 1, 1)
  if quad then
    local imw = iw
    local imh = ih
    for w = 0, fsal_width/imw do
      for h = 0, fsal_height/imh do
        love.graphics.draw(image, quad, w * imw, h * imh)
      end
    end
  else
    local imw = image:getWidth()
    local imh = image:getHeight()
    for w = 0, fsal_width/imw do
      for h = 0, fsal_height/imh do
        love.graphics.draw(image, w * imw, h * imh)
      end
    end
  end
  love.graphics.pop()
end

-- fsal_simpleinit()
--   Initialize FSAL with the current window width and height.
--   The startup state is fixed to window mode.
function fsal_simpleinit()
  fsal_initialize(love.graphics.getWidth(), love.graphics.getHeight(), false)
end

-- fsal_begindraw()
--   Prepare a canvas for FSAL and start drawing on it.
function fsal_begindraw()
  canvas = fsal_canvas()
  fsal_drawstart(canvas)
end

-- fsal_showcanvas()
--   Finish drawing on the canvas for FSAL,
--   and display what you have drawn on the fore screen.
function fsal_showcanvas()
  fsal_drawend()
  fsal_show(canvas)
end

-- fsal_simplechange(key, cur)
--   When the F11 key is pressed,
--   it toggles between windowed and full screen mode.
--   Call this in love.keypressed(key) and pass key directly to this function.
--     key: Parameter "key" at love.keypressed(key).
--     cur: Whether or not the mouse cursor is displayed in full screen.
--       true: show mouse cursor / false: hide mouse cursor
function fsal_simplechange(key, cur)
  if key == "f11" then
    fsal_modechange(cur)
  end
end

-- fsal_getmargininfo()
--   If you want to draw outside the game screen area in full screen mode,
--   you can call this function to get the current screen coordinates.
--     return value: Information of two areas as a table. {<area1>, <area2>}
--       Each area infomation is as {type: Area type, x: Leftmost coordinate,
--       y: Topmost coordinate, width: Area width, height: Area height}
--       Area type: "L":Left side, "R": Right side, "T": Top side, "B": Bottom side
function fsal_getmargininfo()
  local res = {}
  if (fsal_offsetx>0) then
    res = {
      {type="L", x=0, y=0, width=fsal_offsetx*fsal_scale, height=fsal_cheight*fsal_scale},
      {type="R", x=(fsal_offsetx+fsal_cwidth)*fsal_scale, y=0,
        width=fsal_offsetx*fsal_scale, height=fsal_cheight*fsal_scale}
      }
  elseif (fsal_offsety>0) then
    res = {
      {type="T", x=0, y=0, width=fsal_cwidth*fsal_scale, height=fsal_offsety*fsal_scale},
      {type="B", x=0, y=(fsal_offsety+fsal_cheight)*fsal_scale,
        width=fsal_cwidth*fsal_scale, height=fsal_offsety*fsal_scale}
      }
  else
    res = {{type="N", x=0, y=0, width=0, height=0}, {type="N", x=0, y=0, width=0, height=0}}
  end
  return res
end

-- fsal_getscmode()
--   Calling this function returns whether the screen is currently full screen or not.
--     return value: Information on whether or not fullscreen mode
--       true: full screen mode / false: window mode
function fsal_getscmode()
  return fsal_scmode
end

-- fsal_convertRV(x, y)
--   Convert the actual coordinates to virtual coordinates of the resolution set by the game.
--     return value: Adjusted Coordinates
--       tx: X coordinate after conversion
--       ty: Y coordinate after conversion
function fsal_convertRV(x, y)
  local tx, ty = -1, -1
  --print(fsal_offsetx*fsal_scale, fsal_offsety*fsal_scale)
  if (math.floor(fsal_offsetx*fsal_scale + 0.5)<=x) and (x<math.floor((fsal_offsetx+fsal_cwidth)*fsal_scale) + 0.5) then
    if (math.floor(fsal_offsety*fsal_scale + 0.5)<=y) and (y<math.floor((fsal_offsety+fsal_cheight)*fsal_scale) + 0.5) then
      tx = math.floor(((x - (fsal_offsetx*fsal_scale)) / fsal_scale) + 0.5)
      ty = math.floor(((y - (fsal_offsety*fsal_scale)) / fsal_scale) + 0.5)
    end
  end
  return tx, ty
end

-- fsal_convertVR(x, y)
--   Convert the virtual coordinates of the resolution set by the game to actual coordinates.
--     return value: Adjusted Coordinates
--       tx: X coordinate after conversion
--       ty: Y coordinate after conversion
function fsal_convertVR(x, y)
  local tx = math.floor((x + fsal_offsetx) * fsal_scale + 0.5)
  local ty = math.floor((y + fsal_offsety) * fsal_scale + 0.5)
  return tx, ty
end

-- fsal_getscale()
--   Obtain the magnification rate at the time of the call.
--     return value: rate of magnification
function fsal_getscale()
  return fsal_scale
end

-- fsal_resize(width, height)
--   Change the FSAL calculation screen size.
--   Must be called passing w=width and h=height from love.resize(w, h).
--     width : Horizontal size of window (px)
--     height: Vertical size of window (px)
function fsal_resize(width, height)
  fsal_width, fsal_height = width, height
  --print(fsal_width, fsal_height)
  --print(fsal_width.." ".. fsal_height)
  fsal_scalex = fsal_width / fsal_cwidth
  fsal_scaley = fsal_height / fsal_cheight
  if (fsal_scalex<fsal_scaley) then
    fsal_hvmode = false
    fsal_offsetx = 0
    fsal_offsety = ((fsal_height / fsal_scalex) - fsal_cheight) / 2
    fsal_scale = fsal_scalex
  else
    fsal_hvmode = true
    fsal_offsetx = ((fsal_width / fsal_scaley) - fsal_cwidth) / 2
    fsal_offsety = 0
    fsal_scale = fsal_scaley
  end
end

-- fsal_resetwinsize()
--   Reset the window size to the FSAL canvas size.
function fsal_resetwinsize()
  if fsal_scmode then
    fsal_modechange()
  end
  local w, h, f = love.window.getMode()
  w, h = fsal_defaultw, fsal_defaulth
  local w2, h2 = love.graphics.getDimensions()
  if ((w~=w2) or (h~=h2)) then
    love.window.setMode(fsal_defaultw, fsal_defaulth, f)
    fsal_resize(fsal_defaultw, fsal_defaulth)
  end
end

-- fsal_setclipmouse(onoff, grab)
--   Toggles whether or not the mouse cursor is clipped within the game area.
--     onoff: Game area mode flag
--       true: trap mouse cursor in game area / false: free mouse cursor from game area
--     grab : Mouse grab flag
--       true: grab mouse cursor in window / false: free mouse cursor from window
function fsal_setclipmouse(onoff, grab)
  fsal_mcmode = onoff
  if (grab~=nil) then
    --print(grab)
    love.mouse.setGrabbed(grab)
  end
end

-- fsal_clipmouse(x, y)
--   Adjusts the coordinates of the mouse cursor within the game area.
--   x, y can be given as the mouse coordinates from which the calculation is made.
--   If x and y are not given, the mouse coordinates are obtained and calculated.
--   The mouse cursor automatically moves to the correction position.
--   If you simply want the coordinates, use getvirtualmouse(x, y) instead.
--   It is unstable when used in LÖVE mouse-related callbacks.
--   Use getvirtualmouse(x, y) in mouse-related callbacks.
--     x: X-coordinate from which to calculate
--     y: Y-coordinate from which to calculate
--     return value: Coordinates after conversion
--       tx: X coordinate after conversion
--       ty: Y coordinate after conversion
function fsal_clipmouse(x, y)
  local mx, my
  if (x and y) then
    mx, my = x, y
  else
    mx, my = love.mouse.getPosition()
  end
  local tx, ty = mx, my
  if fsal_mcmode then
    local minmx, minmy = fsal_offsetx*fsal_scale, fsal_offsety*fsal_scale
    local maxmx, maxmy = (fsal_offsetx+fsal_cwidth)*fsal_scale, (fsal_offsety+fsal_cheight)*fsal_scale
    if ((mx<minmx) or (maxmx<=mx)) then
      tx = math.min(math.max(minmx, mx), maxmx)
    end
    if ((my<minmy) or (maxmy<=my)) then
      ty = math.min(math.max(minmy, my), maxmy)
    end
    --dx, dy = mx, my
    if ((tx~=mx) or (ty~=my)) then
      love.mouse.setPosition(tx, ty)
    end
  end
  --print(mx, my, "|", minmx, minmy, "|", maxmx, maxmy, "|", dx, dy)
  tx = math.min(math.max(math.floor(tx/fsal_scale-fsal_offsetx+0.5),0), fsal_cwidth-1)
  ty = math.min(math.max(math.floor(ty/fsal_scale-fsal_offsety+0.5),0), fsal_cheight-1)
  return tx, ty
end

-- fsal_getvirtualmouse(x, y)
--   Get mouse coordinates corrected to the game canvas size.
--     x: X-coordinate from which to calculate
--     y: Y-coordinate from which to calculate
--     return value: Coordinates after conversion
--       tx: X coordinate after conversion
--       ty: Y coordinate after conversion
function fsal_getvirtualmouse(x, y)
  local tx, ty = -1, -1
  --print(fsal_offsetx*fsal_scale, fsal_offsety*fsal_scale)
  if (math.floor(fsal_offsetx*fsal_scale)>=x) then
    tx = 0
  elseif (x>=math.floor((fsal_offsetx+fsal_cwidth)*fsal_scale)) then
    tx = fsal_cwidth-1
  else
    tx = math.floor(((x - (fsal_offsetx*fsal_scale)) / fsal_scale))
  end
  if (math.floor(fsal_offsety*fsal_scale)>=y) then
    ty = 0
  elseif (y>=math.floor((fsal_offsety+fsal_cheight)*fsal_scale)) then
    ty = fsal_cheight-1
  else
    ty = math.floor(((y - (fsal_offsety*fsal_scale)) / fsal_scale))
  end
  return tx, ty
end
